from typing import Union

import numpy
import numpy as np
from shapely.geometry import Polygon

def iou_2d(points:Union[list,tuple,numpy.ndarray],other_points:Union[list,tuple,numpy.ndarray]):
    """
    Calculate 2D iou of two polygons made from 2 sets of 2d points
    :param points: in shap of N x 2
    :param other_points: in shap of N x 2
    :return: iou 2d ranging from 0 to 1
    """
    def calculate_iou(box_1, box_2):
        poly_1 = Polygon(box_1)
        poly_2 = Polygon(box_2)
        iou = poly_1.intersection(poly_2).area / poly_1.union(poly_2).area
        return iou
    return calculate_iou(points,other_points)

if __name__=="__main__":
    points = [[0,0],[10,10],[0,10]]
    other_points = [[0, 20], [10, 10], [0, 0]]
    print(iou_2d(points,other_points))


import json
import os
import shutil


def recreate_dir(dir, cancel_if_exist=True):
    if os.path.exists(dir):
        if cancel_if_exist:
            return
        else:
            shutil.rmtree(dir)
    if not os.path.exists(dir):
        os.makedirs(dir)




from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import ModelCheckpoint, BackboneFinetuning
from torch.optim import Optimizer
from torch.utils.data import DataLoader
import yaml
import torch

from .backbones import get_backbone
from .CenterNet import CenterNet

TRAIN_CONFIG = "train_config"
VAL_CONFIG = "val_config"


class CenterNetBackboneFineTuning(BackboneFinetuning):
    def freeze_before_training(self, pl_module: "pl.LightningModule") -> None:
        print("Freeze layer before training")
        backbone = pl_module.backbone
        for name, param in backbone.named_parameters():
            if "deconv_layers" not in name:
                print(f"Freezing {name}")
                param.requires_grad = False

    def finetune_function(
            self, pl_module: "pl.LightningModule", epoch: int, optimizer: Optimizer, opt_idx: int
    ) -> None:
        print(f"Finetune function epoch {epoch}")
        if self.unfreeze_backbone_at_epoch == epoch:
            backbone = pl_module.backbone
            for name, param in backbone.named_parameters():
                if "deconv_layers" not in name:
                    print(f"Unfreezing {name}")
                    param.requires_grad = True


def create_checkpoint_callback(config):
    checkpoint_callback = ModelCheckpoint(
        monitor=config["monitor"],
        dirpath=config["dirpath"],
        filename="checkpoint-{epoch:02d}-{val_loss:.2f}.pth",
        save_top_k=config["save_top_k"],
        mode=config["mode"],
        save_weights_only=True
    )
    return checkpoint_callback


def create_backbone_unfreeze_callback(config):
    unfreeze_bbone_at_epoch = config["unfreeze_bbone_epoch"]
    initial_denom_lr = config["initial_denom_lr"]
    callback = CenterNetBackboneFineTuning(
        unfreeze_bbone_at_epoch, initial_denom_lr=initial_denom_lr
    )
    return callback


def load_model_for_inference(config_path, device, load_pretrained_backbone=True):
    with open(config_path, "r") as config_file:
        yaml_config = yaml.safe_load(config_file)
        model = load_centernet_model_with_config(yaml_config, load_pretrained_backbone).to(device)
        ckpt_path = yaml_config["model"]["ckpt"]
        ckpt = torch.load(ckpt_path, map_location="cpu")
        model.load_state_dict(ckpt["state_dict"])
        model.eval()
        return model


def load_centernet_model_with_config(config, load_bbone_pretrained=True):
    layers = config["model"]["backbone_layers"]
    backbone = get_backbone(layers, load_bbone_pretrained)
    num_classes = config["model"]["num_classes"]
    head_conv_channel = int(config["model"]["head_conv_channel"])
    max_object = int(config["model"]["max_object"])
    input_shape = config["model"]["input_shape"]
    model = CenterNet(backbone, num_classes, head_conv_channel, max_object, input_shape, config[TRAIN_CONFIG])
    return model


class CenterNetTrainer:
    def __init__(self, train_data, val_data, config_path):
        self.train_data = train_data
        self.val_data = val_data
        self.config = yaml.safe_load(open(config_path, "r"))
        self.train_bs = self.config[TRAIN_CONFIG]["batch_size"]
        self.val_bs = self.config[VAL_CONFIG]["batch_size"]
        self.shuffle = self.config[TRAIN_CONFIG]["shuffle"]
        self.num_workers = self.config[TRAIN_CONFIG]["num_workers"]

    def train(self):
        # load model
        model = load_centernet_model_with_config(self.config)

        train_loader = DataLoader(self.train_data, self.train_bs, self.shuffle, num_workers=self.num_workers,
                                  pin_memory=True, drop_last=True)
        val_loader = DataLoader(self.val_data, self.val_bs, num_workers=self.num_workers)

        # create callbacks
        checkpoint_callback = create_checkpoint_callback(self.config[TRAIN_CONFIG]["callback"])
        backbone_unfreeze_callback = create_backbone_unfreeze_callback(self.config[TRAIN_CONFIG])

        # config trainer
        num_gpus = self.config[TRAIN_CONFIG]["gpus"]
        gpus = None if num_gpus == 0 else num_gpus
        #
        trainer = Trainer(gpus=gpus, max_epochs=self.config["train_config"]["epoch"],
                          callbacks=[checkpoint_callback, backbone_unfreeze_callback], amp_backend="apex",
                          amp_level="02",
                          auto_lr_find=True)
        trainer.fit(model, train_loader, val_loader)



import math
import torch.nn as nn
import torch.utils.model_zoo as model_zoo

BN_MOMENTUM = 0.1

model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
}


def conv3x3(in_planes, out_planes, stride=1, padding=1):
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=padding, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes, momentum=BN_MOMENTUM)
        self.conv3 = nn.Conv2d(planes, planes * self.expansion, kernel_size=1,
                               bias=False)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion,
                                  momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


def fill_up_weights(up):
    w = up.weight.data
    f = math.ceil(w.size(2) / 2)
    c = (2 * f - 1 - f % 2) / (2. * f)
    for i in range(w.size(2)):
        for j in range(w.size(3)):
            w[0, 0, i, j] = \
                (1 - math.fabs(i / f - c)) * (1 - math.fabs(j / f - c))
    for c in range(1, w.size(0)):
        w[c, 0, :, :] = w[0, 0, :, :]


def fill_fc_weights(layers):
    for m in layers.modules():
        if isinstance(m, nn.Conv2d):
            nn.init.normal_(m.weight, std=0.001)
            # torch.nn.init.kaiming_normal_(m.weight.data, nonlinearity='relu')
            # torch.nn.init.xavier_normal_(m.weight.data)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)


class ResNetBackbone(nn.Module):

    RESNET_SPEC = {18: (BasicBlock, [2, 2, 2, 2]),
                   34: (BasicBlock, [3, 4, 6, 3]),
                   50: (Bottleneck, [3, 4, 6, 3]),
                   101: (Bottleneck, [3, 4, 23, 3]),
                   152: (Bottleneck, [3, 8, 36, 3])}

    def __init__(self, block, layers):
        self.inplanes = 64
        self.deconv_with_bias = False

        super(ResNetBackbone, self).__init__()
        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)

        # used for deconv layers
        self.deconv_layers = self._make_deconv_layer(
            3,
            [256, 128, 64],
            [4, 4, 4],
        )

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion, momentum=BN_MOMENTUM),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def _get_deconv_cfg(self, deconv_kernel, index):
        if deconv_kernel == 4:
            padding = 1
            output_padding = 0
        elif deconv_kernel == 3:
            padding = 1
            output_padding = 1
        elif deconv_kernel == 2:
            padding = 0
            output_padding = 0

        return deconv_kernel, padding, output_padding

    def _make_deconv_layer(self, num_layers, num_filters, num_kernels):
        assert num_layers == len(num_filters), \
            'ERROR: num_deconv_layers is different len(num_deconv_filters)'
        assert num_layers == len(num_kernels), \
            'ERROR: num_deconv_layers is different len(num_deconv_filters)'

        layers = []
        for i in range(num_layers):
            kernel, padding, output_padding = \
                self._get_deconv_cfg(num_kernels[i], i)

            planes = num_filters[i]
            # fc = DCN(self.inplanes, planes,
            #          kernel_size=(3, 3), stride=1,
            #          padding=1, dilation=1, deformable_groups=1)
            fc = nn.Conv2d(self.inplanes, planes,
                           kernel_size=3, stride=1,
                           padding=1, dilation=1, bias=False)
            fill_fc_weights(fc)
            up = nn.ConvTranspose2d(
                in_channels=planes,
                out_channels=planes,
                kernel_size=kernel,
                stride=2,
                padding=padding,
                output_padding=output_padding,
                bias=self.deconv_with_bias)
            fill_up_weights(up)

            layers.append(fc)
            layers.append(nn.BatchNorm2d(planes, momentum=BN_MOMENTUM))
            layers.append(nn.ReLU(inplace=True))
            layers.append(up)
            layers.append(nn.BatchNorm2d(planes, momentum=BN_MOMENTUM))
            layers.append(nn.ReLU(inplace=True))
            self.inplanes = planes

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.deconv_layers(x)
        return x

    def init_weights(self, num_layers):
        url = model_urls['resnet{}'.format(num_layers)]
        pretrained_state_dict = model_zoo.load_url(url)
        print('=> loading pretrained model {}'.format(url))
        self.load_state_dict(pretrained_state_dict, strict=False)
        print('=> init deconv weights from normal distribution')
        for name, m in self.deconv_layers.named_modules():
            if isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)


class PoseResNetBackbone(ResNetBackbone):
    def __init__(self, block, layers,heads, head_conv):
        super(PoseResNetBackbone, self).__init__(block, layers)
        self.heads = heads
        for head in self.heads:
            classes = self.heads[head]
            if head_conv > 0:
                fc = nn.Sequential(
                    nn.Conv2d(64, head_conv,
                              kernel_size=3, padding=1, bias=True),
                    nn.ReLU(inplace=True),
                    nn.Conv2d(head_conv, classes,
                              kernel_size=1, stride=1,
                              padding=0, bias=True))
                if 'heatmap' in head:
                    fc[-1].bias.data.fill_(-2.19)
                else:
                    fill_fc_weights(fc)
            else:
                fc = nn.Conv2d(64, classes,
                               kernel_size=1, stride=1,
                               padding=0, bias=True)
                if 'heatmap' in head:
                    fc.bias.data.fill_(-2.19)
                else:
                    fill_fc_weights(fc)
            self.__setattr__(head, fc)

    def forward(self, x):
        x = super(PoseResNetBackbone, self).forward(x)
        ret = {}
        for head in self.heads:
            ret[head] = self.__getattr__(head)(x)
        return ret


def get_backbone(num_layers,load_bbone_pretrained=True):
    block_class, layers = ResNetBackbone.RESNET_SPEC[num_layers]
    model = ResNetBackbone(block_class, layers)
    if load_bbone_pretrained:
        model.init_weights(num_layers)
    return model


def get_pose_net(num_layers, heads, head_conv=256):
    block_class, layers = ResNetBackbone.RESNET_SPEC[num_layers]

    model = PoseResNetBackbone(block_class, layers, heads, head_conv=head_conv)
    model.init_weights(num_layers)
    return model



import numpy as np
from torch import sigmoid
from torch.nn import MaxPool2d
import torch
import cv2

from .. import CenterNet


def decode_centernet(ct_hm, ct_offset_map, reg_map):
    bs, cn, hw, hh = ct_hm.shape

    max_pooling = MaxPool2d(kernel_size=3, stride=1, padding=1)

    max_per_channels = ct_hm == ct_hm.view(bs, cn, -1).max(dim=2, keepdim=True).values.unsqueeze(-1)
    positives = (ct_hm == max_pooling(ct_hm)) * (ct_hm >= 0.3) * (
            ct_hm == ct_hm.max(dim=1, keepdim=True).values.repeat(1, ct_hm.shape[1], 1, 1)) * max_per_channels
    pos_locs = positives.nonzero()

    batch_indices = pos_locs[:, 0:1]
    class_ids = pos_locs[:, 1]
    ct_offset = ct_offset_map[pos_locs[:, 0], :, pos_locs[:, 2], pos_locs[:, 3]]
    score = ct_hm[pos_locs[:, 0], pos_locs[:, 1], pos_locs[:, 2], pos_locs[:, 3]].unsqueeze(dim=1)
    wh = reg_map[pos_locs[:, 0], :, pos_locs[:, 2], pos_locs[:, 3]]
    return torch.cat((batch_indices, class_ids.unsqueeze(dim=1), pos_locs[:, 2].unsqueeze(dim=1),
                      pos_locs[:, 3].unsqueeze(dim=1), ct_offset, wh, score), dim=1)


def parse_output_centernet(pred):
    ct_hm = pred["hm"]
    offset = sigmoid(pred["offset"])
    reg = pred["reg"]
    boxes = decode_centernet(ct_hm, offset, reg)
    return boxes


@torch.no_grad()
def infer(model: CenterNet, image: np.ndarray, device="cpu"):
    h, w, c = image.shape()
    ih, iw = model.input_shape
    down_scale_h = h / ih
    down_scale_w = w / iw
    image = cv2.resize(image, dsize=(iw, ih), interpolation=cv2.INTER_LINEAR)
    image = np.transpose(image, (2, 0, 1))
    image = torch.from_numpy(image).to(device)
    image = torch.unsqueeze(image, 0)
    output = model(image)
    batch_output = parse_output_centernet(output)

    res = []
    for detection in batch_output:
        image_index, class_id, yint, xint, offsetx, offsety, w, h, score = detection
        class_id = int(class_id)
        yint = int(yint)
        xint = int(xint)
        offsetx = float(offsetx)
        offsety = float(offsety)
        w = float(w) * down_scale_w
        h = float(h) * down_scale_h
        score = round(float(score), 2)
        xc = (xint + offsetx) * down_scale_w * model.down_ratio_x
        yc = (yint + offsety) * down_scale_h * model.down_ratio_y
        res.append([xc, yc, w, h, class_id, score])
    return res
